math.randomseed(os.time())

function genPrim(width, height, no_frame) 
-- if no_frame is true, the algorithm will not leave a wall around the maze
-- no_frame only has an effect on mazes with an even number of rows or columns
  -- initialize the map with all walls --
  local map = {}
  local in_the_maze = {} -- temporary map with additional info
  for w=1, width do
    map[w] = {}
    in_the_maze[w] = {}
    for h=1, height do
      map[w][h] = true
      in_the_maze[w][h] = false
    end
  end


  local walls = {} -- the wall list

  -- helper function 
  local function add_walls (rx, ry) -- add walls around a given cell to the list
    -- I am saving the coordinates to the wall but also to the "cell on the opposite
    -- side", which is crucial for this algorithm.
    if (rx > 1) and not in_the_maze[rx-1][ry] and map[rx-1][ry] then
      table.insert(walls, {x=rx-1, y=ry, op_x=rx-2, op_y=ry})
    end
    if (rx < width) and not in_the_maze[rx+1][ry] and map[rx+1][ry] then
      table.insert(walls, {x=rx+1, y=ry, op_x=rx+2, op_y=ry})
    end
    if (ry > 1) and not in_the_maze[rx][ry-1] and map[rx][ry-1] then
      table.insert(walls, {x=rx, y=ry-1, op_x=rx, op_y=ry-2})
    end
    if (ry < height) and not in_the_maze[rx][ry+1] and map[rx][ry+1] then
      table.insert(walls, {x=rx, y=ry+1, op_x=rx, op_y=ry+2})
    end
  end

  if not no_frame then
    width = width - 1
    height = height - 1
  end

  -- start of the algorithm --
  -- pick a cell at random --
  local rx = math.random(1, math.floor(width/2)) * 2
  local ry = math.random(1, math.floor(height/2)) * 2
  -- mark it as part of the maze --
  in_the_maze[rx][ry] = true
  map[rx][ry] = false
  -- add the walls of the cell to the wall list --
  add_walls(rx, ry)

  -- while there are walls in the list --
  while #walls > 0 do
    -- pick a random wall from the list
    local current_wall = math.random(#walls)
    local cw = walls[current_wall]

    -- is the cell on the opposite side in the maze or out of bounds?
    if (cw.op_x < 1) or (cw.op_x > width) or (cw.op_y < 1) or (cw.op_y > height)
    or in_the_maze[cw.op_x][cw.op_y] then
      -- then remove the wall from the list --
      table.remove(walls, current_wall)
    else
      -- make the wall a passage --
      map[cw.x][cw.y] = false
      in_the_maze[cw.x][cw.y] = true

      -- add cell on the opposite side to the maze --
      map[cw.op_x][cw.op_y] = false
      in_the_maze[cw.op_x][cw.op_y] = true
      -- add walls  to the list --			
      add_walls(cw.op_x, cw.op_y)
    end
  end
  return map
end

function tablesum(T)
  local sum = 0
  for _,n in ipairs(T) do if n ~= nil and type(n) == "number" then sum = sum + n end end
  return sum
end
function tablelength(T)
  return #T
end
function lastof(T)
  return T[tablelength(T)]
end
function randElement(arr)
  return arr[math.random(1,#arr)]
end

local size = 20

local freeTiles = {}

local map = genPrim(size,size)
for y=1,size do
  for x=1,size do
    if type(map[y][x]) == "boolean" then
      if map[y][x] then
        map[y][x] = -1
      else
        map[y][x] = 0
        table.insert(freeTiles,{x=x,y=y})
      end
    end
  end
end

local mazestart = randElement(freeTiles)
local mazeend = randElement(freeTiles)
map[mazestart.y][mazestart.x] = 1

local path = {{x=mazeend.x,y=mazeend.y}}
local x = lastof(path).x
local y = lastof(path).y

local pathView = {{x=mazeend.x,y=mazeend.y}}

local blockSize = love.graphics.getWidth()/size

local showNums = false

local solved = false

function newMaze(s)
  size = s
  freeTiles = {}
  map = genPrim(size,size)
  for y=1,size do
    for x=1,size do
      if type(map[y][x]) == "boolean" then
        if map[y][x] then
          map[y][x] = -1
        else
          map[y][x] = 0
          table.insert(freeTiles,{x=x,y=y})
        end
      end
    end
  end
  mazestart = randElement(freeTiles)
  mazeend = randElement(freeTiles)
  map[mazestart.y][mazestart.x] = 1
  path = {{x=mazeend.x,y=mazeend.y}}
  x = lastof(path).x
  y = lastof(path).y
  pathView = {{x=mazeend.x,y=mazeend.y}}
  blockSize = love.graphics.getWidth()/size
  solved = false
end

function checkFill()
  local toAdd = {}
  for y=1,size do
    for x=1,size do
      if map[y][x] > 0 then
        if map[y][x+1] == 0 and x+1 < size then table.insert(toAdd,{x+1,y,map[y][x]+1}) end
        if map[y][x-1] == 0 and x-1 > 0 then table.insert(toAdd,{x-1,y,map[y][x]+1}) end
        if y+1 < size and map[y+1][x] == 0 then table.insert(toAdd,{x,y+1,map[y][x]+1}) end
        if y-1 > 0 and map[y-1][x] == 0 then table.insert(toAdd,{x,y-1,map[y][x]+1}) end
      end
    end
  end
  for _,t in ipairs(toAdd) do
    map[t[2]][t[1]] = t[3]
  end
end

function solvePath()
  repeat
    local args = {}
    if map[y][x+1] > 0 then table.insert(args,map[y][x+1]) else table.insert(args,tablesum(map)) end
    if map[y][x-1] > 0 then table.insert(args,map[y][x-1]) else table.insert(args,tablesum(map)) end
    if map[y+1][x] > 0 then table.insert(args,map[y+1][x]) else table.insert(args,tablesum(map)) end
    if map[y-1][x] > 0 then table.insert(args,map[y-1][x]) else table.insert(args,tablesum(map)) end
    for i,s in ipairs(args) do
      if s == nil or s < 1 then args[i] = tablesum(args) + 1 end
    end
    local tochoose = math.min(args[1],args[2],args[3],args[4])
    if tochoose == map[y][x+1] then x = x + 1 
    elseif tochoose == map[y][x-1] then x = x - 1 
    elseif tochoose == map[y+1][x] then y = y + 1 
    elseif tochoose == map[y-1][x] then y = y - 1 end
    table.insert(path,{x=x,y=y})
  until x == mazestart.x and y == mazestart.y
end

function touchesEnd()
  return map[mazeend.y][mazeend.x] > 0
end

function completeSolve()
  repeat checkFill() until touchesEnd()
  solvePath()
end

function love.keypressed(key)
  if key == 'space' and not solved then
    completeSolve()
    solved = true
  end
  if key == 'd' then showNums = not showNums end
  if key == 'n' then
    newMaze(size)
  end
  if key == 'w' then newMaze(size+1) end
  if key == 's' and size > 5 then newMaze(size-1) end
end

function love.draw()
  for y=1,size do
    for x=1,size do
      if map[y][x] == -1 then
        love.graphics.setColor(100,100,100)
        love.graphics.rectangle("fill",(x-1)*blockSize,(y-1)*blockSize,blockSize,blockSize)
      end
    end
  end
  if solved then
    if #pathView ~= #path then
      table.insert(pathView,path[#path-#pathView])
    end
    for _, piece in ipairs(pathView) do
      love.graphics.setColor(0,255,255,100)
      love.graphics.rectangle("fill",(piece.x-1)*blockSize,(piece.y-1)*blockSize,blockSize,blockSize)
    end
  end
  love.graphics.setColor(0,255,255)
  love.graphics.rectangle("fill",(mazestart.x-1)*blockSize,(mazestart.y-1)*blockSize,blockSize,blockSize)
  love.graphics.setColor(255,255,0)
  love.graphics.rectangle("fill",(mazeend.x-1)*blockSize,(mazeend.y-1)*blockSize,blockSize,blockSize)
  love.graphics.setColor(255,255,255)
  if showNums then
    for y=1,size do
      for x=1,size do
        love.graphics.print(map[y][x],(x-1)*blockSize,(y-1)*blockSize)
      end
    end
  end
  love.graphics.print("space - solve\nd - show numbers\nn - new maze\nw - increase size (new maze)\ns - decrease size (new maze)\nsize = " .. size)
end